"""Project constants"""
import os

DEFAULT_INPUT_SHAPE = 684, 1024
MODEL_ENCODER_NAME = os.getenv("MODEL_ENCODER_NAME", "mobile_sam.encoder.onnx")
MODEL_DECODER_NAME = os.getenv("MODEL_DECODER_NAME", "mobile_sam.decoder.onnx")



import json
import logging
import os
from pathlib import Path


def stats_pathname(pathname: Path | str):
    current_pathname = Path(pathname)
    return current_pathname.is_dir()


def create_folder_if_not_exists(pathname: Path | str):
    """Create a folder given its path.

    Args:
        pathname: folder Path or string

    Returns:

    """
    current_pathname = Path(pathname)
    try:
        print(f"Pathname exists? {current_pathname.exists()}, That's a folder? {current_pathname.is_dir()}...")
        logging.info(f"Pathname exists? {current_pathname.exists()}, That's a folder? {current_pathname.is_dir()}...")
        current_pathname.unlink(missing_ok=True)
    except PermissionError as pe:
        print(f"permission denied on removing pathname before folder creation:{pe}.")
        logging.error(f"permission denied on removing pathname before folder creation:{pe}.")
    except IsADirectoryError as errdir:
        print(f"that's a directory:{errdir}.")
        logging.error(f"that's a directory:{errdir}.")

    print(f"Creating pathname: {current_pathname} ...")
    logging.info(f"Creating pathname: {current_pathname} ...")
    current_pathname.mkdir(mode=0o770, parents=True, exist_ok=True)

    print(f"assertion: pathname exists and is a folder: {current_pathname} ...")
    logging.info(f"assertion: pathname exists and is a folder: {current_pathname} ...")
    assert current_pathname.is_dir()


def folders_creation(folders_map: dict | str = None, ignore_errors: bool = True):
    """Create all folders listed within the folders_map argument (this argument can be a dict or a json string).
    If folders_map is None the function will try to load the 'FOLDERS_MAP' env variable, then will load that json into
    dict. Once loaded and parsed the folders_map variable, the function will loop over the dict to create the folders
    using the `create_folder_if_not_exists()` function.

    Args:
        folders_map: dict or string map of folder string
        ignore_errors: bool needed to eventually ignore errors on folder creation

    Returns:

    """
    enforce_validation_with_getenv = folders_map is None
    if enforce_validation_with_getenv:
        folders_map = os.getenv("FOLDERS_MAP")
    try:
        folders_dict = folders_map if isinstance(folders_map, dict) else json.loads(folders_map)
        for folder_env_ref, folder_env_path in folders_dict.items():
            logging.info(f"folder_env_ref:{folder_env_ref}, folder_env_path:{folder_env_path}.")
            create_folder_if_not_exists(folder_env_path)
            print("========")
            if enforce_validation_with_getenv:
                assert os.getenv(folder_env_ref) == folder_env_path
    except (json.JSONDecodeError, TypeError) as jde:
        logging.error(f"jde:{jde}.")
        msg = "double check your variables, e.g. for misspelling like 'FOLDER_MAP'"
        msg += "instead than 'FOLDERS_MAP', or invalid json values."
        logging.error(msg)
        for k_env, v_env in dict(os.environ).items():
            logging.info(f"{k_env}, v_env:{v_env}.")
        if not ignore_errors:
            raise TypeError(jde)


if __name__ == '__main__':
    folders_creation()




"""custom type hints"""
from enum import Enum
from typing import TypedDict, NewType

from PIL.Image import Image
from numpy import ndarray


class ListStr(list[str]): pass


class DictStrInt(dict[str, int]): pass


class DictStr(dict[str]): pass


class DictStrAny(dict[str, any]): pass


class ListDict(list[dict]): pass


class ListFloat(list[float]): pass


class ListInt(list[int]): pass


class TupleInt(tuple[int]): pass


TupleInt2 = NewType("TupleInt", tuple[int, int])


class TupleNdarrayInt(tuple[ndarray, int]): pass


class TupleNdarrayFloat(tuple[ndarray, float]): pass


class LlistFloat(ListFloat): pass


class TupleFloat(tuple[float]): pass


class TupleFloatAny(tuple[float, any]): pass


PIL_Image = Image


class StrEnum(str, Enum):
    pass


class EmbeddingImage(TypedDict):
    image_embedding: ndarray
    original_size: TupleInt
    transform_matrix: ndarray


class EmbeddingPILImage(TypedDict):
    image_embedding: ndarray
    original_size: TupleInt2
    resized_size: TupleInt2


class MatplotlibBackend(StrEnum):
    gtk3agg = "gtk3agg"
    gtk3cairo = "gtk3cairo"
    gtk4agg = "gtk4agg"
    gtk4cairo = "gtk4cairo"
    macosx = "macosx"
    nbagg = "nbagg"
    notebook = "notebook"
    qtagg = "qtagg"
    qtcairo = "qtcairo"
    qt5agg = "qt5agg"
    qt5cairo = "qt5cairo"
    tkagg = "tkagg"
    tkcairo = "tkcairo"
    webagg = "webagg"
    wx = "wx"
    wxagg = "wxagg"
    wxcairo = "wxcairo"
    agg = "agg"
    cairo = "cairo"
    pdf = "pdf"
    pgf = "pgf"
    ps = "ps"
    svg = "svg"
    template = "template"


EmbeddingDict = dict[str, EmbeddingImage]
EmbeddingPILDict = dict[str, EmbeddingPILImage]



"""Get machine learning predictions from geodata raster images"""
import os
from pathlib import Path

import structlog
from dotenv import load_dotenv

from samgis_core.utilities import session_logger


load_dotenv()
PROJECT_ROOT_FOLDER = Path(globals().get("__file__", "./_")).absolute().parent.parent
PROJECT_MODEL_FOLDER = Path(PROJECT_ROOT_FOLDER / "machine_learning_models")
MODEL_FOLDER = os.getenv("MODEL_FOLDER", PROJECT_MODEL_FOLDER)
LOG_JSON_FORMAT = bool(os.getenv("LOG_JSON_FORMAT", False))
log_level = os.getenv("LOG_LEVEL", "INFO")
session_logger.setup_logging(json_logs=LOG_JSON_FORMAT, log_level=log_level)
app_logger = structlog.stdlib.get_logger(__name__)



"""Serialize objects"""
from typing import Mapping

from samgis_core import app_logger


def serialize(obj: any, include_none: bool = False):
    """
    Return the input object into a serializable one

    Args:
        obj: Object to serialize
        include_none: bool to indicate if include also keys with None values during dict serialization

    Returns:
        serialized object
    """
    return _serialize(obj, include_none)


def _serialize(obj: any, include_none: bool):
    from numpy import ndarray as np_ndarray, floating as np_floating, integer as np_integer

    primitive = (int, float, str, bool)
    # print(type(obj))
    try:
        if obj is None:
            return None
        elif isinstance(obj, np_integer):
            return int(obj)
        elif isinstance(obj, np_floating):
            return float(obj)
        elif isinstance(obj, np_ndarray):
            return obj.tolist()
        elif isinstance(obj, primitive):
            return obj
        elif type(obj) is list:
            return _serialize_list(obj, include_none)
        elif type(obj) is tuple:
            return list(obj)
        elif type(obj) is bytes:
            return _serialize_bytes(obj)
        elif isinstance(obj, Exception):
            return _serialize_exception(obj)
        # elif isinstance(obj, object):
        #     return _serialize_object(obj, include_none)
        else:
            return _serialize_object(obj, include_none)
    except Exception as e_serialize:
        app_logger.error(f"e_serialize::{e_serialize}, type_obj:{type(obj)}, obj:{obj}.")
        return f"object_name:{str(obj)}__object_type_str:{str(type(obj))}."


def _serialize_object(obj: Mapping[any, object], include_none: bool) -> dict[any]:
    from bson import ObjectId

    res = {}
    if type(obj) is not dict:
        keys = [i for i in obj.__dict__.keys() if (getattr(obj, i) is not None) or include_none]
    else:
        keys = [i for i in obj.keys() if (obj[i] is not None) or include_none]
    for key in keys:
        if type(obj) is not dict:
            res[key] = _serialize(getattr(obj, key), include_none)
        elif isinstance(obj[key], ObjectId):
            continue
        else:
            res[key] = _serialize(obj[key], include_none)
    return res


def _serialize_list(ls: list, include_none: bool) -> list:
    return [_serialize(elem, include_none) for elem in ls]


def _serialize_bytes(b: bytes) -> dict[str, str]:
    import base64
    encoded = base64.b64encode(b)
    return {"value": encoded.decode('ascii'), "type": "bytes"}


def _serialize_exception(e: Exception) -> dict[str, str]:
    return {"msg": str(e), "type": str(type(e)), **e.__dict__}



"""
Define a machine learning model executed by ONNX Runtime (https://onnxruntime.ai/)
for Segment Anything (https://segment-anything.com).
Modified from
- https://github.com/vietanhdev/samexporter/
- https://github.com/AndreyGermanov/sam_onnx_full_export/

Copyright (c) 2023 Viet Anh Nguyen, Andrey Germanov
Copyright (c) 2024-today Alessandro Trinca Tornidor

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from numpy import array as np_array, pad as np_pad, zeros, ndarray, concatenate, float32
from onnxruntime import get_available_providers, InferenceSession

from samgis_core import app_logger
from samgis_core.utilities.constants import DEFAULT_INPUT_SHAPE
from samgis_core.utilities.type_hints import ListDict, EmbeddingPILImage, PIL_Image
from samgis_core.utilities.utilities import convert_ndarray_to_pil, apply_coords


class SegmentAnythingONNX2:
    """
    Segmentation model using Segment Anything.
    Compatible with onnxruntime 1.17.x and later
    """

    def __init__(self, encoder_model_path: str, decoder_model_path: str) -> None:
        self.target_size = DEFAULT_INPUT_SHAPE[1]
        self.input_size = DEFAULT_INPUT_SHAPE

        # Load models
        providers = get_available_providers()

        # Pop TensorRT Runtime due to crashing issues
        # TODO: Add back when TensorRT backend is stable
        providers = [p for p in providers if p != "TensorrtExecutionProvider"]

        if providers:
            app_logger.info(
                "Available providers for ONNXRuntime: %s", ", ".join(providers)
            )
        else:
            app_logger.warning("No available providers for ONNXRuntime")
        self.encoder_session = InferenceSession(
            encoder_model_path, providers=providers
        )
        self.encoder_input_name = self.encoder_session.get_inputs()[0].name
        app_logger.info(f"encoder_input_name:{self.encoder_input_name}.")
        self.decoder_session = InferenceSession(
            decoder_model_path, providers=providers
        )

    @staticmethod
    def get_input_points(prompt: ListDict) -> tuple[ndarray]:
        """
        Get input points from a prompt dict list.

        Args:
            prompt: dict list

        Returns:
            tuple of points, labels ndarray ready for Segment Anything inference

        """
        points = []
        labels = []
        for mark in prompt:
            if mark["type"] == "point":
                points.append(mark["data"])
                labels.append(mark["label"])
            elif mark["type"] == "rectangle":
                points.append([mark["data"][0], mark["data"][1]])  # top left
                points.append(
                    [mark["data"][2], mark["data"][3]]
                )  # bottom right
                labels.append(2)
                labels.append(3)
        points, labels = np_array(points), np_array(labels)
        return points, labels

    def encode(self, img: PIL_Image | ndarray) -> EmbeddingPILImage:
        """
        Calculate embedding and metadata for a single image.

        Args:
            img: input image to embed

        Returns:
            embedding image dict useful to store and cache image embeddings

        """
        resized_image = self.preprocess_image(img)
        padded_input_tensor = self.padding_tensor(resized_image)

        # 2. GET IMAGE EMBEDDINGS USING IMAGE ENCODER (`size` argument here is like ndarray `shape`)
        outputs = self.encoder_session.run(None, {"images": padded_input_tensor})
        image_embedding = outputs[0]
        img = convert_ndarray_to_pil(img)
        return {
            "image_embedding": image_embedding,
            "original_size": img.size,
            "resized_size": resized_image.size
        }

    def predict_masks(self, embedding: EmbeddingPILImage, prompt: ListDict) -> ndarray:
        """
        Predict masks for a single image.

        Args:
            embedding: input image embedding dict
            prompt: Segment Anything input prompt

        Returns:
            prediction masks ndarray; this should have (1, 1, **image.shape) shape

        """
        input_points, input_labels = self.get_input_points(prompt)

        # Add a batch index, concatenate a padding point, and transform.
        onnx_coord = concatenate([input_points, np_array([[0.0, 0.0]])], axis=0)[None, :, :]
        onnx_label = concatenate([input_labels, np_array([-1])], axis=0)[None, :].astype(float32)

        onnx_coord = apply_coords(onnx_coord, embedding)
        orig_width, orig_height = embedding["original_size"]
        app_logger.info(f"onnx_coord:{onnx_coord}.")

        # RUN DECODER TO GET MASK
        onnx_mask_input = zeros((1, 1, 256, 256), dtype=float32)
        onnx_has_mask_input = zeros(1, dtype=float32)
        output_masks, _, _ = self.decoder_session.run(None, {
            "image_embeddings": embedding["image_embedding"],
            "point_coords": onnx_coord,
            "point_labels": onnx_label,
            "mask_input": onnx_mask_input,
            "has_mask_input": onnx_has_mask_input,
            "orig_im_size": np_array([orig_height, orig_width], dtype=float32)
        })
        return output_masks

    def preprocess_image(self, img: PIL_Image | ndarray) -> ndarray:
        """
        Resize image preserving aspect ratio using `output_size_target` as a long side.

        Args:
            img: input ndarray/PIL image

        Returns:
            image ndarray

        """
        from PIL import Image

        app_logger.info(f"image type:{type(img)}, shape/size:{img.size}.")
        try:
            orig_width, orig_height = img.size
        except TypeError:
            img = Image.fromarray(img)
            orig_width, orig_height = img.size

        resized_height = self.target_size
        resized_width = int(self.target_size / orig_height * orig_width)

        if orig_width > orig_height:
            resized_width = self.target_size
            resized_height = int(self.target_size / orig_width * orig_height)

        img = img.resize((resized_width, resized_height), Image.Resampling.BILINEAR)
        return img

    def padding_tensor(self, img: PIL_Image | ndarray) -> ndarray:
        """
        Pad an image ndarray/tensor to given instance self.target_size

        Args:
            img: input ndarray/PIL image

        Returns:
            image ndarray

        """
        # Prepare input tensor from image
        tensor_input = np_array(img)
        resized_width, resized_height = img.size

        # Normalize input tensor numbers
        mean = np_array([123.675, 116.28, 103.53])
        std = np_array([[58.395, 57.12, 57.375]])
        tensor_input = (tensor_input - mean) / std

        # Transpose input tensor to shape (Batch,Channels,Height,Width
        tensor_input = tensor_input.transpose(2, 0, 1)[None, :, :, :].astype(float32)

        # Make image square self.target_size x self.target_size by padding short side by zeros
        tensor_input = np_pad(tensor_input, ((0, 0), (0, 0), (0, 0), (0, self.target_size - resized_width)))
        if resized_height < resized_width:
            tensor_input = np_pad(tensor_input, ((0, 0), (0, 0), (0, self.target_size - resized_height), (0, 0)))

        return tensor_input



import structlog.stdlib
from numpy import ndarray
from matplotlib import pyplot as plt

from samgis_core.utilities.type_hints import ListStr, MatplotlibBackend

FigAxes = tuple[plt.Figure, plt.Axes]


logger = structlog.stdlib.get_logger(__file__)


def helper_imshow_output_expected(
        img_list: list[ndarray], titles_list: ListStr, cmap: str = "gist_rainbow", plot_size: int = 5,
        show=False, debug: bool = False, close_after: float = 0.0) -> FigAxes:
    """
    Simple way to display a list of images with their titles, color map.
    Should work also in an automate environments, like tests (use a `close_after` argument > 0)

    Args:
        img_list: ndarray images to display
        titles_list: title images
        cmap: color map
        plot_size: figure plot size
        show: fire plt.show() action if needed
        debug: workaround useful in an interactive context, like Pycharm debugger
        close_after: close after give seconds (useful in tests, contrasted to 'debug' option)

    Returns:
        tuple of matplotlib Figure, Axes

    """
    n = len(img_list)
    assert len(titles_list) == n
    fig, ax = plt.subplot_mosaic([
        titles_list
    ], figsize=(n * plot_size, plot_size))

    for title, img in zip(titles_list, img_list):
        ax[title].imshow(img, cmap=cmap)
        ax[title].legend()
    if show:
        if debug:
            plt.pause(0.01)
            plt.show()
        if close_after > 0:
            plt.pause(close_after)
            plt.show(block=False)
            plt.close("all")
    return fig, ax


def imshow_raster(
        raster, title, cmap: str = "gist_rainbow", interpolation: str = None, alpha=None, transform=None, plot_size=5,
        show=False, debug: bool = False, close_after: float = 0.0, backend: MatplotlibBackend = None) -> FigAxes:
    """
    Displays raster images lists/arrays with titles, legend, alpha transparency, figure sizes
    and geographic transformations, if not none (leveraging rasterio.plot)

    Args:
        raster: image to display
        title: title image
        cmap: color map
        interpolation: interpolation type
        alpha: alpha transparency
        transform: geographic transform, eventually used for map representation by rasterio
        plot_size: figure plot size
        show: fire plt.show() action if needed
        debug: workaround useful in an interactive context, like Pycharm debugger
        close_after: close after give seconds (useful in tests, contrasted to 'debug' option)
        backend: matplotlib backend string

    Returns:
        tuple of matplotlib Figure, Axes

    """
    from rasterio import plot

    if not backend:
        backend = plt.get_backend()
        plt.rcParams["backend"] = backend
    logger.info(f"use {backend} as matplotlib backend...")

    fig, ax = plt.subplots(figsize=(plot_size, plot_size))
    raster_ax = raster[0] if transform is not None else raster
    image_hidden = ax.imshow(raster_ax, cmap=cmap, interpolation=interpolation, alpha=alpha)
    if transform is not None:
        plot.show(raster, transform=transform, ax=ax, cmap=cmap, interpolation=interpolation, alpha=alpha)
    fig.colorbar(image_hidden, ax=ax)
    ax.set_title(title)
    if show:
        if debug:
            plt.pause(0.01)
            plt.show()
        if close_after > 0:
            plt.pause(close_after)
            plt.show(block=False)
            plt.close("all")
    return fig, ax



import logging
import sys

import structlog
from structlog.types import EventDict, Processor


# https://github.com/hynek/structlog/issues/35#issuecomment-591321744
def rename_event_key(_, __, event_dict: EventDict) -> EventDict:
    """
    Log entries keep the text message in the `event` field, but Datadog
    uses the `message` field. This processor moves the value from one field to
    the other.
    See https://github.com/hynek/structlog/issues/35#issuecomment-591321744
    """
    event_dict["message"] = event_dict.pop("event")
    return event_dict


def drop_color_message_key(_, __, event_dict: EventDict) -> EventDict:
    """
    Uvicorn logs the message a second time in the extra `color_message`, but we don't
    need it. This processor drops the key from the event dict if it exists.
    """
    event_dict.pop("color_message", None)
    return event_dict


def setup_logging(json_logs: bool = False, log_level: str = "INFO"):
    """Enhance the configuration of structlog.
    Needed for correlation id injection with fastapi middleware in samgis-web.
    After the use of logging_middleware() in samgis_web.web.middlewares, add also the CorrelationIdMiddleware from
    'asgi_correlation_id' package. (See 'tests/web/test_middlewares.py' in samgis_web).
    To change an input parameter like the log level, re-run the function changing the parameter
    (no need to re-instantiate the logger instance: it's a hot change)

    Args:
        json_logs: set logs in json format
        log_level: log level string

    Returns:

    """
    timestamper = structlog.processors.TimeStamper(fmt="iso")

    shared_processors: list[Processor] = [
        structlog.contextvars.merge_contextvars,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.stdlib.ExtraAdder(),
        drop_color_message_key,
        timestamper,
        structlog.processors.StackInfoRenderer(),
        # adapted from https://www.structlog.org/en/stable/standard-library.html
        # If the "exc_info" key in the event dict is either true or a
        # sys.exc_info() tuple, remove "exc_info" and render the exception
        # with traceback into the "exception" key.
        structlog.processors.format_exc_info,
        # If some value is in bytes, decode it to a Unicode str.
        structlog.processors.UnicodeDecoder(),
        # Add callsite parameters.
        structlog.processors.CallsiteParameterAdder(
            {
                structlog.processors.CallsiteParameter.FUNC_NAME,
                structlog.processors.CallsiteParameter.LINENO,
            }
        ),
        # Render the final event dict as JSON.
    ]

    if json_logs:
        # We rename the `event` key to `message` only in JSON logs, as Datadog looks for the
        # `message` key but the pretty ConsoleRenderer looks for `event`
        shared_processors.append(rename_event_key)
        # Format the exception only for JSON logs, as we want to pretty-print them when
        # using the ConsoleRenderer
        shared_processors.append(structlog.processors.format_exc_info)

    structlog.configure(
        processors=shared_processors
        + [
            # Prepare event dict for `ProcessorFormatter`.
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )

    log_renderer: structlog.types.Processor
    if json_logs:
        log_renderer = structlog.processors.JSONRenderer()
    else:
        log_renderer = structlog.dev.ConsoleRenderer()

    formatter = structlog.stdlib.ProcessorFormatter(
        # These run ONLY on `logging` entries that do NOT originate within
        # structlog.
        foreign_pre_chain=shared_processors,
        # These run on ALL entries after the pre_chain is done.
        processors=[
            # Remove _record & _from_structlog.
            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
            log_renderer,
        ],
    )

    handler = logging.StreamHandler()
    # Use OUR `ProcessorFormatter` to format all `logging` entries.
    handler.setFormatter(formatter)
    root_logger = logging.getLogger()
    root_logger.addHandler(handler)
    root_logger.setLevel(log_level.upper())

    for _log in ["uvicorn", "uvicorn.error"]:
        # Clear the log handlers for uvicorn loggers, and enable propagation
        # so the messages are caught by our root logger and formatted correctly
        # by structlog
        logging.getLogger(_log).handlers.clear()
        logging.getLogger(_log).propagate = True

    # Since we re-create the access logs ourselves, to add all information
    # in the structured log (see the `logging_middleware` in main.py), we clear
    # the handlers and prevent the logs to propagate to a logger higher up in the
    # hierarchy (effectively rendering them silent).
    logging.getLogger("uvicorn.access").handlers.clear()
    logging.getLogger("uvicorn.access").propagate = False

    def handle_exception(exc_type, exc_value, exc_traceback):
        """
        Log any uncaught exception instead of letting it be printed by Python
        (but leave KeyboardInterrupt untouched to allow users to Ctrl+C to stop)
        See https://stackoverflow.com/a/16993115/3641865
        """
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return

        root_logger.error(
            "Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback)
        )

    sys.excepthook = handle_exception



"""Various utilities (logger, time benchmark, args dump, numerical and stats info)"""
from copy import deepcopy

import numpy as np
from numpy import ndarray, float32

from samgis_core import app_logger
from samgis_core.utilities.serialize import serialize
from samgis_core.utilities.type_hints import EmbeddingPILImage, PIL_Image


def _prepare_base64_input(sb):
    if isinstance(sb, str):
        # If there's any unicode here, an exception will be thrown and the function will return false
        return bytes(sb, 'ascii')
    elif isinstance(sb, bytes):
        return sb
    raise ValueError("Argument must be string or bytes")


def _is_base64(sb: str | bytes):
    import base64

    try:
        sb_bytes = _prepare_base64_input(sb)
        decoded = base64.b64decode(sb_bytes, validate=True)
        return base64.b64encode(decoded).decode("utf-8") == sb_bytes.decode("utf-8")
    except ValueError:
        return False


def base64_decode(s):
    """
    Decode base64 strings

    Args:
        s: input string

    Returns:
        decoded string
    """
    import base64

    if isinstance(s, str) and _is_base64(s):
        return base64.b64decode(s, validate=True).decode("utf-8")

    return s


def base64_encode(sb: str | bytes) -> bytes:
    """
    Encode input strings or bytes as base64

    Args:
        sb: input string or bytes

    Returns:
        base64 encoded bytes
    """
    import base64

    sb_bytes = _prepare_base64_input(sb)
    return base64.b64encode(sb_bytes)


def hash_calculate(arr) -> str | bytes:
    """
    Return computed hash from input variable (typically a numpy array).

    Args:
        arr: input variable

    Returns:
        computed hash from input variable
    """
    from hashlib import sha256
    from base64 import b64encode
    from numpy import ndarray as np_ndarray

    if isinstance(arr, np_ndarray):
        hash_fn = sha256(arr.data)
    elif isinstance(arr, dict):
        import json

        serialized = serialize(arr)
        variable_to_hash = json.dumps(serialized, sort_keys=True).encode('utf-8')
        hash_fn = sha256(variable_to_hash)
    elif isinstance(arr, str):
        try:
            hash_fn = sha256(arr)
        except TypeError:
            app_logger.warning(f"TypeError, re-try encoding arg:{arr},type:{type(arr)}.")
            hash_fn = sha256(arr.encode('utf-8'))
    elif isinstance(arr, bytes):
        hash_fn = sha256(arr)
    else:
        raise ValueError(f"variable 'arr':{arr} of type '{type(arr)}' not yet handled.")
    return b64encode(hash_fn.digest())


def convert_ndarray_to_pil(pil_image: PIL_Image | ndarray):
    """
    Check if an image is a ndarray and then convert to a PIL Image instance.

    Args:
        pil_image: PIL image or ndarray

    Returns:
        PIL Image

    """
    from PIL import Image

    if isinstance(pil_image, ndarray):
        pil_image = Image.fromarray(pil_image)
    return pil_image


def apply_coords(coords: ndarray, embedding: EmbeddingPILImage):
    """
    Expects a numpy np_array of length 2 in the final dimension. Requires the
    original image size in (H, W) format.

        Args:
            coords: coordinates ndarray
            embedding: PIL image embedding dict

        Returns:
            coordinates ndarray

    """
    orig_width, orig_height = embedding["original_size"]
    resized_width, resized_height = embedding["resized_size"]
    coords = deepcopy(coords).astype(float)

    coords[..., 0] = coords[..., 0] * (resized_width / orig_width)
    coords[..., 1] = coords[..., 1] * (resized_height / orig_height)

    return coords.astype(float32)


def normalize_array(arr: ndarray, new_h: int | float = 255., type_normalization: str = "int") -> ndarray:
    """
    Normalize numpy array between 0 and 'new_h' value. Default dtype of output array is int


    Args:
        arr: input numpy array
        new_h: max value of output array
        type_normalization: default dtype of output array

    Returns:
        numpy array
    """
    arr = arr.astype(float)
    arr_max = np.nanmax(arr)
    arr_min = np.nanmin(arr)
    scaled_arr = (arr - arr_min) / (arr_max - arr_min)
    multiplied_arr = scaled_arr * new_h
    return multiplied_arr.astype(int) if type_normalization == "int" else multiplied_arr



import os
import subprocess
from pathlib import Path

from dotenv import load_dotenv

from samgis_core import app_logger
from samgis_core.utilities.type_hints import ListStr


load_dotenv()
root_folder = Path(globals().get("__file__", "./_")).absolute().parent.parent.parent
env_project_root_folder = os.getenv("PROJECT_ROOT_FOLDER", str(root_folder))
env_input_css_path = os.getenv("INPUT_CSS_PATH")


def assert_envs(envs_list: ListStr) -> None:
    """
    Assert env variables are not empty.

    Args:
        envs_list: list of env variables

    Returns:

    """
    for current_env in envs_list:
        try:
            assert current_env is not None and current_env != ""
        except AssertionError as aex:
            app_logger.error(f"error on assertion for current_env: {current_env}.")
            raise aex


def read_std_out_err(std_out_err: str, output_type: str, command: ListStr) -> None:
    """
    Capture the standard output or standard error content for a given system command pipe.

    Args:
        std_out_err: str
        output_type: output type (stdout or stderr)
        command: command executed

    Returns:

    """
    output = std_out_err.split("\n")
    app_logger.info(f"output type:{output_type} for command:{' '.join(command)}.")
    for line in iter(output):
        app_logger.info(f"output_content_home stdout:{line.strip()}.")
    app_logger.info("########")


def run_command(commands_list: ListStr, capture_output: bool = True, text: bool = True, check: bool = True) -> None:
    """
    Run a system command and capture its output.
    See https://docs.python.org/3.11/library/subprocess.html#subprocess.run for more details

    Args:
        commands_list: list of single string commands
        capture_output: if true, stdout and stderr will be captured
        text: if true, capture stdout and stderr as strings
        check: If check is true, and the process exits with a non-zero exit code, a CalledProcessError exception will
               be raised. Attributes of that exception hold the arguments, the exit code, and stdout and stderr if they
               were captured.

    Returns:

    """
    try:
        output_content_home = subprocess.run(
            commands_list,
            capture_output=capture_output,
            text=text,
            check=check
        )
        read_std_out_err(output_content_home.stdout, "stdout", commands_list)
        read_std_out_err(output_content_home.stderr, "stderr", commands_list)
    except Exception as ex:
        app_logger.error(f"ex:{ex}.")
        raise ex


def build_frontend(
        project_root_folder: str | Path,
        input_css_path: str | Path,
        output_dist_folder: Path = root_folder / "static" / "dist",
        index_page_filename: str = "index.html",
        output_css_filename: str = "output.css",
        force_build: bool = False,
    ) -> bool:
    """
    Build a static [Vue js](https://vuejs.org/), [tailwindcss](https://tailwindcss.com/) frontend.
    If force_build is False, the function also check if index_page_filename and output_css_filename already exists:
    in this case skip the build.

    Args:
        project_root_folder: Project folder that contains the static frontend
        input_css_path: file path pointing to the input css file
        output_dist_folder: dist folder path where to write the frontend bundle
        index_page_filename: index html filename
        output_css_filename: output css filename
        force_build: useful to skip the frontend build

    Returns:
        state of the build (True in case of build completed, False in case of build skipped)

    """
    assert_envs([
        str(project_root_folder),
        str(input_css_path)
    ])
    project_root_folder = Path(project_root_folder)
    index_html_pathfile = Path(output_dist_folder) / index_page_filename
    output_css_pathfile = Path(output_dist_folder) / output_css_filename
    if not force_build and output_css_pathfile.is_file() and index_html_pathfile.is_file():
        app_logger.info("frontend ok, build_frontend not necessary...")
        return False

    # install deps
    os.chdir(project_root_folder / "static")
    current_folder = os.getcwd()
    app_logger.info(f"current_folder:{current_folder}, install pnpm...")
    run_command(["which", "npm"])
    run_command(["npm", "install", "-g", "npm", "pnpm"])
    app_logger.info(f"install pnpm dependencies...")
    run_command(["pnpm", "install"])

    # build frontend dist and assert for its correct build
    output_css = str(output_dist_folder / output_css_filename)
    output_index_html = str(output_dist_folder / index_page_filename)
    output_dist_folder = str(output_dist_folder)
    app_logger.info(f"pnpm: build '{output_dist_folder}'...")
    run_command(["pnpm", "build"])
    app_logger.info(f"pnpm: ls -l {output_index_html}:")
    run_command(["ls", "-l", output_index_html])
    cmd = ["pnpm", "tailwindcss", "-i", str(input_css_path), "-o", output_css]
    app_logger.info(f"pnpm: {' '.join(cmd)}...")
    run_command(["pnpm", "tailwindcss", "-i", str(input_css_path), "-o", output_css])
    app_logger.info(f"pnpm: ls -l {output_css}:")
    run_command(["ls", "-l", output_css])
    app_logger.info(f"end!")
    return True


if __name__ == '__main__':
    build_frontend(
        project_root_folder=Path(env_project_root_folder),
        input_css_path=Path(env_input_css_path)
    )



from numpy import array as np_array, uint8, zeros, ndarray

from samgis_core import app_logger, MODEL_FOLDER
from samgis_core.prediction_api.sam_onnx2 import SegmentAnythingONNX2
from samgis_core.utilities.constants import MODEL_ENCODER_NAME, MODEL_DECODER_NAME
from samgis_core.utilities.type_hints import ListDict, PIL_Image, TupleNdarrayInt, EmbeddingPILDict


def get_raster_inference(
        img: PIL_Image | ndarray, prompt: ListDict, models_instance: SegmentAnythingONNX2, model_name: str
) -> TupleNdarrayInt:
    """
    Get inference output for a given image using a SegmentAnythingONNX model

    Args:
        img: input PIL Image
        prompt: list of prompt dict
        models_instance: SegmentAnythingONNX instance model
        model_name: model name string

    Returns:
        raster prediction mask, prediction number

    """
    np_img = np_array(img)
    app_logger.info(f"img type {type(np_img)}, prompt:{prompt}.")
    app_logger.debug(f"onnxruntime input shape/size (shape if PIL) {np_img.size}.")
    try:
        app_logger.debug(f"onnxruntime input shape (NUMPY) {np_img.shape}.")
    except Exception as e_shape:
        app_logger.error(f"e_shape:{e_shape}.")
    app_logger.info(f"instantiated model {model_name}, ENCODER {MODEL_ENCODER_NAME}, "
                    f"DECODER {MODEL_DECODER_NAME} from {MODEL_FOLDER}: Creating embedding...")
    embedding = models_instance.encode(np_img)
    app_logger.debug(f"embedding created, running predict_masks with prompt {prompt}...")
    return get_raster_inference_using_existing_embedding(embedding, prompt, models_instance)


def get_inference_embedding(
        img: PIL_Image | ndarray, models_instance: SegmentAnythingONNX2, model_name: str, embedding_key: str,
        embedding_dict: EmbeddingPILDict) -> EmbeddingPILDict:
    """add an embedding to the embedding dict if needed

    Args:
        img: input PIL Image
        models_instance: SegmentAnythingONNX instance model
        model_name: model name string
        embedding_key: embedding id
        embedding_dict: embedding dict object

    Returns:
        raster dict

    """
    if embedding_key in embedding_dict:
        app_logger.info("found embedding in dict...")
    if embedding_key not in embedding_dict:
        np_img = np_array(img)
        app_logger.info(f"prepare embedding using img type {type(np_img)}.")
        app_logger.debug(f"onnxruntime input shape/size (shape if PIL) {np_img.size}.")
        try:
            app_logger.debug(f"onnxruntime input shape (NUMPY) {np_img.shape}.")
        except Exception as e_shape:
            app_logger.error(f"e_shape:{e_shape}.")
        app_logger.info(f"instantiated model {model_name}, ENCODER {MODEL_ENCODER_NAME}, "
                        f"DECODER {MODEL_DECODER_NAME} from {MODEL_FOLDER}: Creating embedding...")
        embedding = models_instance.encode(np_img)
        embedding_dict[embedding_key] = embedding
    return embedding_dict


def get_raster_inference_using_existing_embedding(
        embedding: dict, prompt: ListDict, models_instance: SegmentAnythingONNX2, folder_write_tmp_on_disk: str = None,
        key: str = None) -> TupleNdarrayInt:
    """
    Get inference output for a given image using a SegmentAnythingONNX model, using an existing embedding instead of a
    new ndarray or PIL image

    Args:
        embedding: dict
        prompt: list of prompt dict
        models_instance: SegmentAnythingONNX instance model
        folder_write_tmp_on_disk: output folder where to write debug images
        key: embedding key

    Returns:
        raster prediction mask, prediction number

    """
    app_logger.info(f"using existing embedding of type {type(embedding)}.")
    inference_out = models_instance.predict_masks(embedding, prompt)
    len_inference_out = len(inference_out[0, :, :, :])
    app_logger.info(f"Created {len_inference_out} prediction_masks,"
                    f"shape:{inference_out.shape}, dtype:{inference_out.dtype}.")
    mask = zeros((inference_out.shape[2], inference_out.shape[3]), dtype=uint8)
    write_tmp_img = bool(folder_write_tmp_on_disk)
    for n, m in enumerate(inference_out[0, :, :, :]):
        app_logger.debug(f"{n}th of prediction_masks shape {inference_out.shape}"
                         f" => mask shape:{mask.shape}, {mask.dtype}.")
        mask[m > 0.0] = 255
        if write_tmp_img:
            from pathlib import Path
            from datetime import datetime
            from samgis_core.utilities.utilities import convert_ndarray_to_pil, normalize_array
            m_normalized = normalize_array(m, type_normalization="float")
            m_out = convert_ndarray_to_pil(m_normalized)
            now = datetime.now().isoformat()
            if len(m.shape) == 2:
                m_out = m_out.convert("L")
            m_out.save(Path(folder_write_tmp_on_disk) / f"mask_{key}_{now}_n{n}.png")
    return mask, len_inference_out


def get_raster_inference_with_embedding_from_dict(
        img: PIL_Image | ndarray, prompt: ListDict, models_instance: SegmentAnythingONNX2, model_name: str,
        embedding_key: str, embedding_dict: dict, folder_write_tmp_on_disk: str = None) -> TupleNdarrayInt:
    """
    Get inference output using a SegmentAnythingONNX model, but get the image embedding from the given embedding dict
     instead of creating a new embedding. This function needs the img argument to update the embedding dict if necessary

    Args:
        img: input PIL Image
        prompt: list of prompt dict
        models_instance: SegmentAnythingONNX instance model
        model_name: model name string
        embedding_key: embedding id
        embedding_dict: embedding images dict
        folder_write_tmp_on_disk: output folder where to write debug images

    Returns:
        raster prediction mask, prediction number

    """
    app_logger.info(f"handling embedding using key {embedding_key}.")
    embedding_dict = get_inference_embedding(img, models_instance, model_name, embedding_key, embedding_dict)
    app_logger.info(f"getting embedding with key {embedding_key} from dict...")
    embedding = embedding_dict[embedding_key]
    n_keys = len(embedding_dict)
    app_logger.info(f"embedding created ({n_keys} keys in embedding dict), running predict_masks with prompt {prompt}.")
    return get_raster_inference_using_existing_embedding(
        embedding, prompt, models_instance, folder_write_tmp_on_disk=folder_write_tmp_on_disk, key=embedding_key)
